/**	Reference table generator, HP48 Utility
  Borland C++ 3.0
  Detlef Mueller, 24.04.1992
  Q&D hack - Usage: GENTAB <[entries file] >[table file]
  
  000 24.04.1992 DM	coded
  100 25.08.1992 DM	modified output to feed it into sasm only
  200 26.10.1994 MH	modified for unix
  300 19.11.1994 MH	modified to make a library suitable for Jazz
  400 03.01.1995 MH	added the DIS.TAB 00000 terminator
  500 19.06.1995 MH	made to output proper headers
  			made preferrance clear via indexing
  600 31.05.1996 MH	added MAXLEN definition to reduce memory use
			changed to use unsigned int for ENTRY->index
  700 23.07.2010 HD now uses extable format; now also lists
			duplicate entries if they exist
  **/

#include	<stdio.h>
#include	<stdlib.h>
#include	<stdarg.h>
#include	<string.h>
#include	<ctype.h>
#include	<errno.h>


#ifdef MSDOS
#include	<alloc.h>
#include	<io.h>			/* MesS-DOS */
#include	<fcntl.h>		/* MesS-DOS */
#endif

#define	MAJVER		1
#define	MINVER		0
#define	REVER		0
#define	WRKFILE		"________.TMP"
#define	STARTCH		' '

#define MAXLEN		63		/* Max name length in chars */

typedef enum
{
	ERROR, WARNING
}
ERRTYPE ;

typedef	struct
{
	char name[MAXLEN+1] ;
	unsigned long addr ;
	unsigned int index ;		/* Added for DIS.TAB generation */
	unsigned int hash ;
	unsigned int length ;
}
ENTRY ;


static ENTRY **table = NULL ;

static unsigned int	entries = 0;
static unsigned int	longest = 0;
static unsigned int	hash[128 - STARTCH] = { 0 } ;

void  Error ( ERRTYPE messt, char *msg, ... )
{
	va_list	argptr ;
	char	*tstr, str[81] ;
	
	va_start( argptr, msg ) ;
	vsprintf( str, msg, argptr ) ;
	
	switch ( messt )
	{
	case ERROR :
		tstr = "\nError: %s " ;
		break ;
		
	case WARNING :
		tstr = "\nWarning: %s " ;
		break ;
		
	default :
		tstr = "Unknown error: %s " ;
	}
	
	fprintf( stderr, tstr, str ) ;

#ifdef MSDOS
	if ( errno != EZERO )
#else
		if ( errno != 0 )		/* No idea what EZERO is.. */
#endif
			perror( "" ) ;
		else
			fputc( '\n', stderr ) ;
	
	va_end( argptr ) ;
	
	if ( messt == ERROR )
		exit( -1 ) ;
}

/* alphabetical sort; shorter names come first */
static int compare ( const void *s1, const void *s2 )
{
	register char	*s_1, *s_2 ;
	
	for ( s_1 = (*(ENTRY **)s1)->name, s_2 = (*(ENTRY **)s2)->name ;
	     *s_1 || *s_2 ; ++s_1, ++s_2 )
		if ( *s_1 != *s_2 )
			return ( *s_1 - *s_2 ) ;
	
	return ( 0 ) ;
}

/*
 * qsort is not stable, but we need stability to get preferrance
 * Thus we assign indexes to the entries based on input order, and if
 * the addresses of the entries match then input order determines
 * comparison result. Thus address sort never returns equality.
 */
static int compare2 ( const void *s1, const void *s2 )
{
	long	ind1, ind2 ;

	if( (*(ENTRY **)s1)->addr > (*(ENTRY **)s2)->addr)
		return 1;
	else if( (*(ENTRY **)s1)->addr < (*(ENTRY **)s2)->addr)
 		return -1;
 	else if( (*(ENTRY **)s1)->index > (*(ENTRY **)s2)->index)
		return 1;
	else
		return -1;
}


/* qsort by hash, then name length, then name; we get stability
 * by again using the index, although the index is only used
 * if we have duplicate entries (of the exact same name)
 */
static int compare3 ( const void *s1, const void *s2 )
{
	register char *s_1, *s_2 ;

	if( (*(ENTRY **)s1)->hash > (*(ENTRY **)s2)->hash )
		return 1;
	else if( (*(ENTRY **)s1)->hash < (*(ENTRY **)s2)->hash )
		return -1;
	else if( (*(ENTRY **)s1)->length > (*(ENTRY **)s2)->length )
		return 1;
	else if( (*(ENTRY **)s1)->length < (*(ENTRY **)s2)->length )
		return -1;
	else
	{
		for ( s_1 = (*(ENTRY **)s1)->name, s_2 = (*(ENTRY **)s2)->name ;
		     *s_1 || *s_2 ; ++s_1, ++s_2 )
			if ( *s_1 != *s_2 )
				return ( *s_1 - *s_2 ) ;
				
		// same hash, same length, same name?! go by index
		fprintf( stderr, "\n\tDuplicate entry: %s", (*(ENTRY **)s1)->name  );
		if( (*(ENTRY **)s1)->index >  (*(ENTRY **)s2)->index )
			return 1;
		else
			return -1;
	}
}


static void Work ( void )
{
	static char	image[] = "\\|/-" ;
	static int	i = 0 ;
	
	fputc( '\b', stderr ) ;
	fputc( image[i = (i + 1) & 3], stderr ) ;
}

static void WriteRTab ( FILE *fout )
{
	unsigned int	i, *h ;
	char		ch = '\0' ;
	ENTRY		**entry ;
	
	fputs( "\nWriting hash table ...", stderr ) ;
	fputs( "RplTab\n", fout );	
	fputs( "\nWriting reference table ...  ", stderr ) ;
	
	for ( i = 0, entry = table ; i < entries ; ++i, ++entry )
	{

		fprintf( fout, "lb%04X\t\tCON(5)\t#%lX\n",
			(*entry)->index, (*entry)->addr );
		fprintf( fout, "lb%04X_len\tCON(2)\t#%lX\n",
			(*entry)->index, (*entry)->length );
		fprintf( fout, "\t\tNIBASC\t`%s`\n\n", (*entry)->name );
		
		if ( ! (i % 100) )
			Work() ;
	}
	
	fputs( "\t\tCON(7)\t0\nRplEnd\n", fout ) ;
	fputs( "\b ", stderr ) ;
}

static void WriteDTab ( FILE *fout )
{
	unsigned int	i;
	ENTRY		**entry ;
	
	fputs( "\nWriting address table ...", stderr ) ;
	fprintf( fout, "\t\tCON(5)\t%ld\t\t* Total number of entries\n", entries);


	fputs( "\t\tREL(5)\tRplTab\t\t* Offset to entries\n", fout );
		fputs( "hash_values\n", fout );

	for (i = 0; i < 128; i++ )
	{
		fprintf( fout, "\t\tREL(5)\thash_%02X\n", i );
	}

	fputs( "DisTab\n", fout );
	for ( i = 0, entry = table ; i < entries ; ++i, ++entry )
	{
		fprintf( fout, "\t\tREL(5)\tlb%04X\t\t%s\n",
			(*entry)->index, (*entry)->name);
		
		if ( ! (i % 100) )
			Work() ;
	}
	fputs( "DisEnd\n", fout ) ;
	fputs( "\b ", stderr ) ;
}

static void WriteDTab2 ( FILE *fout )
{
	unsigned int	i, j;
	ENTRY		**entry ;


	fputs( "\nWriting hash offset table ...", stderr ) ;
	fputs( "hash_tables\nhash_00\n", fout ) ;

	for (i = 0, j = 0, entry = table ; i < entries ; ++i, ++entry )
	{
		if( (*entry)->hash != j )
		{
			j++;
			fprintf( fout, "\t\tCON(5)\t0\nhash_%02X\n", j );
		}
		fprintf( fout, "\t\tREL(5)\tlb%04X_len\t\t%s\n",
			(*entry)->index, (*entry)->name );	
	}
	
	fputs( "\t\tCON(5)\t0\n", fout );

}


static void ReadTab ( FILE *fin )
{
	static char	line[80], buffer[256] ;
	char		*pl, *pa ;
	char		*strname;
	unsigned int	i ;
	long		addr ;
	ENTRY		entry, **tab ;
	FILE		*ftmp ;

	if ( ! (ftmp = fopen( WRKFILE, "wb" )) )
		Error( ERROR, "Can't create <%s>", WRKFILE ) ;

	fprintf( stderr, "\nReading and converting input ...  " ) ;
	
	while ( fgets( line, 80, fin ) )
		if ( strlen( line ) && *line != '\n' )
		{
			pl = strcpy( buffer, line ) ;
			
			if (    ! (pl = strtok( *pl == '=' ? pl + 1 : pl, "\n\r\t " ))
			    || strlen( pl ) > MAXLEN
			    || *pl == '*'
			    || *pl > '~'
			    || ! strtok( NULL, "\n\r\t " )
			    || ! (pa = strtok( NULL, "\n\r\t " ))
			    || ! (i = sscanf( *pa == '#' ? pa + 1 : pa, "%lX", &addr ))
			    || i == EOF )
				continue ;
				

			/* ignore romptrs */
//			if (*pl == '~')
//				continue;

			/* flash pointers */
			if (*pl == '^')
				addr = (addr & 0x0000000FL) | ((addr & 0x0FFFF000L) >> 8);
				
			/* ensure after adjustments that addr is 5 nibbles */
			if (addr & 0xFFF00000L)
				continue;
				
			/* in case we want alpha hash */
			hash[*pl - STARTCH] = 1 ;
			
			strcpy( entry.name, pl ) ;
			entry.addr = addr ;
			entry.index = 0 ;
			entry.length = strlen( entry.name );
			if (entry.length > longest) longest = entry.length;
			strname = entry.name;
			
			/* hash = sum of char nums; masked with #7Fh */
			entry.hash = 0;
			for ( i = 0; i < entry.length; i++ )
			{
			entry.hash += *strname;
			strname++;
			}
			entry.hash = entry.hash & 0x0000007FL ;
			
			if ( ! fwrite( &entry, sizeof( ENTRY ), 1, ftmp ) )
				Error( ERROR, "Write error on <%s>", WRKFILE ) ;
			
			if ( ! (++entries % 100) )
				Work() ;
		}
	
	fclose( ftmp ) ;
	
	if ( ! entries )
		Error( ERROR, "Sorry, no entries found" ) ;
	
	fputs( "\b ", stderr ) ;
	fprintf( stderr, "\nFound %u symbols", entries ) ;
	fprintf( stderr, "\nLongest entry: %u chars\nGenerating table ...  ", longest ) ;
	
	if ( ! (ftmp = fopen( WRKFILE, "rb" )) )
		Error( ERROR, "Can't reopen <%s>", WRKFILE ) ;
	
	if ( ! (table = (ENTRY **)malloc( entries * sizeof( ENTRY *) )) )
		Error( ERROR, "Can't alloc room for %u entries", entries ) ;
	
	for ( i = 0, tab = table ; i < entries ; ++i, ++tab )
	{
		if ( ! (*tab = (ENTRY *)malloc( sizeof( ENTRY ) )) )
			Error( ERROR, "Can't alloc %d bytes room for %uth entry",
			      sizeof( ENTRY ), i ) ;
		
		if ( ! fread( *tab, sizeof( ENTRY ), 1, ftmp ) )
			Error( ERROR, "Can't read %uth entry from <%s>", i, WRKFILE ) ;
		if ( ! (i % 100) )
			Work() ;
	}
	
	fputs( "\b ", stderr ) ;
	
	fclose( ftmp ) ;
	unlink( WRKFILE ) ;
}

void  main ( int argc, char *argv[] )
{
	int	i;
	ENTRY	**tab ;
#ifdef MSDOS
	fprintf( stderr,
		"\n%s, Version %d.%d.%d, %s  %s  BC %X.%X, (c) 1992 DM.\n",
		argv[0], MAJVER, MINVER, REVER, __DATE__, __TIME__,
		__TURBOC__ >> 8, __TURBOC__ & 0x00FF ) ;	/* MesS-DOS */
	setmode( fileno( stdout ), O_BINARY ) ;		/* MesS-DOS */
	setmode( fileno( stdin ), O_BINARY ) ;		/* MesS-DOS */
#endif

	ReadTab( stdin ) ;
	/* Output download header */
	fputs( "\t\tNIBASC\t\\HPHP49-X\\\n", stdout ) ;
	fputs( "\t\tCON(5)\t#02BEE\t\tDOEXT3\n", stdout );
	fputs( "\t\tREL(5)\t->tableend\n", stdout );


	/* Assign preferrance indexes based on input order */
	for( i = 0, tab = table ; i < entries ; ++i, ++tab)
		(*tab)->index = i;


	/* Output DIS.TAB */
	fprintf( stderr, "\nSorting (by address)..." ) ;
	qsort( table, entries, sizeof( ENTRY * ), compare2 ) ;
	WriteDTab( stdout ) ;

	/* output offset to hash tables */
	fprintf( stderr, "\nSorting (by hash/length/name)..." ) ;
	qsort( table, entries, sizeof( ENTRY * ), compare3 ) ;
	WriteDTab2( stdout ) ;
	
	/* Output RPL.TAB */
	fprintf( stderr, "\nSorting (by name)..." ) ;
	qsort( table, entries, sizeof( ENTRY * ), compare ) ;
	WriteRTab( stdout ) ;

	/* end of entries table */
	fputs( "->tableend\n", stdout );

	free( table ) ;
	fprintf( stderr, "\nTo create library table execute:\n");
	fprintf( stderr, "\tsasm -EHN tab\n");
	exit( 0 ) ;
}
